package as.adamsmith.etherealdialpad.dsp;

import android.app.Service;
import android.content.Intent;
import android.content.SharedPreferences;
import android.os.IBinder;
import android.os.Process;
import android.preference.PreferenceManager;
import android.util.Log;

public class SynthService extends Service {
	private static final String TAG = "SynthService";
	
	Thread thread;
	
	@Override
	public IBinder onBind(Intent intent) {
		final SharedPreferences synthPrefs = PreferenceManager
			.getDefaultSharedPreferences(getApplicationContext());
		
		final float[] scale = buildScale(synthPrefs.getString("quantizer", "1,4,6,9,11"));
		final int octaves = Integer.parseInt(synthPrefs.getString("octaves", "3"));
		
		final boolean delay = synthPrefs.getBoolean("delay", true);
		final boolean flange = synthPrefs.getBoolean("flange", false);
		final boolean softTimbre = synthPrefs.getBoolean("softTimbre", true);
		final boolean softEnvelope = synthPrefs.getBoolean("softEnvelope", true);
		final boolean duet = synthPrefs.getBoolean("duet", true);
    	
		final WtOsc ugOscA1 = new WtOsc();
		final WtOsc ugOscA2 = new WtOsc();
		final ExpEnv ugEnvA = new ExpEnv();
	    
		final WtOsc ugOscB1 = new WtOsc();
		final WtOsc ugOscB2 = new WtOsc();
		final ExpEnv ugEnvB = new ExpEnv();
	    
		if(softTimbre) {
			ugOscA1.fillWithHardSin(7.0f);
			ugOscB1.fillWithHardSin(7.0f);
			ugOscA2.fillWithHardSin(2.0f);
			ugOscB2.fillWithHardSin(2.0f);
		} else {
			ugOscA1.fillWithSqrDuty(0.6f);
			ugOscB1.fillWithSqrDuty(0.6f);
			ugOscA2.fillWithSqrDuty(0.9f);
			ugOscB2.fillWithSqrDuty(0.9f);
		}
		
		final Dac ugDac = new Dac();
		final Delay ugDelay = new Delay(UGen.SAMPLE_RATE/2);
		final Flange ugFlange = new Flange(UGen.SAMPLE_RATE/64,0.25f);
		
	    if(delay) {
	    	
	    	ugEnvA.chuck(ugDelay);
	    	ugEnvB.chuck(ugDelay);
	    	
	    	if(flange) {
	    		ugDelay.chuck(ugFlange).chuck(ugDac);
	    	} else {
	    		ugDelay.chuck(ugDac);
	    	}
	    } else {
	    	if(flange) {
	    		ugEnvA.chuck(ugFlange);
	    		ugEnvB.chuck(ugFlange);
	    		ugFlange.chuck(ugDac);
	    	} else {
	    		ugEnvA.chuck(ugDac);
	    		ugEnvB.chuck(ugDac);
	    	}
	    }
	    
	    
	    ugOscA1.chuck(ugEnvA);
	    ugOscB1.chuck(ugEnvB);
		ugOscA2.chuck(ugEnvA);
	    ugOscB2.chuck(ugEnvB);
	    
	    if(!softEnvelope) {
	    	ugEnvA.setFactor(ExpEnv.hardFactor);
	    	ugEnvB.setFactor(ExpEnv.hardFactor);
	    }
		
		if(thread != null) {
			Log.e(TAG,
					"Someone tried to bind when there was already an audio thread!",
					new RuntimeException());
		}
		
		thread = new Thread() {
			@Override
			public void run() {
				Log.d(TAG, "started audio rendering");
				
				System.gc();
				Process.setThreadPriority(Process.THREAD_PRIORITY_AUDIO);
		        
				ugDac.open();
		        
				while(!isInterrupted()) {
					ugDac.tick();
				}
				
				ugDac.close();
				Log.d(TAG, "finished audio rendering");
			}
		};
		
		thread.start();
		
		return new ISynthService.Stub() {
			
			@Override
			public float[] getScale() {
				return scale;
			}
			
			@Override
			public boolean isDuet() {
				return duet;
			}
			
			@Override
			public int getOctaves() {
				return octaves;
			}
			
			@Override public void primaryOff() {
				ugEnvA.setActive(false);
			}

			@Override public void primaryOn() {				
				ugEnvA.setActive(true);
			}

			@Override public void primaryXY(float x, float y) {
				ugOscA1.setFreq(buildFrequency(scale, octaves, x));
				if(duet) {
					ugOscA2.setFreq(buildFrequency(scale, octaves, y));
				} else {
					ugEnvA.setGain(2*y*y);
				}
			}

			@Override public void secondaryOff() {
				ugEnvB.setActive(false);
			}

			@Override public void secondaryOn() {
				ugEnvB.setActive(true);
			}

			@Override public void secondaryXY(float x, float y) {
				ugOscB1.setFreq(buildFrequency(scale, octaves, x));
				if(duet) {
					ugOscB2.setFreq(buildFrequency(scale, octaves, y));
				} else {
					ugEnvB.setGain(2*y*y);
				}
			}
		}; 
	}
	
	
	@Override
	public void onDestroy() {
		thread.interrupt();
		try {
			thread.join();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
	
	
	static float[] buildScale(String quantizerString) {
		if(quantizerString != null && quantizerString.length() > 0) {
			String[] parts = quantizerString.split(",");
			float[] scale = new float[parts.length];
			for(int i = 0; i < parts.length; i++) {
				scale[i] = Float.parseFloat(parts[i]);
			}
			return scale;
		} else {
			return null;
		}
	}
	
	static float buildFrequency(final float[] scale, final int octaves, float input) {
		input = Math.min(Math.max(input, 0.0f), 1.0f);
		final float base = 48;
		
		float mapped;
		if(scale == null) {
			mapped = base + input*octaves*12.0f;
		} else {
			int idx = (int)((scale.length*octaves+1)*input);
			mapped = base + scale[idx%scale.length] + 12*(idx/scale.length);
		}
		return (float)Math.pow(2, (mapped-69.0f)/12.0f) * 440.0f;
		
	}
}
